# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from urllib import parse

from oslo_utils import encodeutils

from novaclient import api_versions
from novaclient.tests.unit.fixture_data import base


class V1(base.Fixture):

    base_url = 'os-hypervisors'
    api_version = '2.1'
    hyper_id_1 = 1234
    hyper_id_2 = 5678
    service_id_1 = 1
    service_id_2 = 2

    @staticmethod
    def _transform_hypervisor_details(hypervisor):
        """Transform a detailed hypervisor view from 2.53 to 2.88."""
        del hypervisor['current_workload']
        del hypervisor['disk_available_least']
        del hypervisor['free_ram_mb']
        del hypervisor['free_disk_gb']
        del hypervisor['local_gb']
        del hypervisor['local_gb_used']
        del hypervisor['memory_mb']
        del hypervisor['memory_mb_used']
        del hypervisor['running_vms']
        del hypervisor['vcpus']
        del hypervisor['vcpus_used']
        hypervisor['uptime'] = 'fake uptime'

    def setUp(self):
        super(V1, self).setUp()

        api_version = api_versions.APIVersion(self.api_version)

        get_os_hypervisors = {
            'hypervisors': [
                {
                    'id': self.hyper_id_1,
                    'hypervisor_hostname': 'hyper1',
                    'state': 'up',
                    'status': 'enabled',
                },
                {
                    'id': self.hyper_id_2,
                    'hypervisor_hostname': 'hyper2',
                    'state': 'up',
                    'status': 'enabled',
                },
            ]
        }

        self.headers = self.json_headers

        self.requests_mock.get(self.url(),
                               json=get_os_hypervisors,
                               headers=self.headers)

        get_os_hypervisors_detail = {
            'hypervisors': [
                {
                    'id': self.hyper_id_1,
                    'service': {
                        'id': self.service_id_1,
                        'host': 'compute1',
                    },
                    'vcpus': 4,
                    'memory_mb': 10 * 1024,
                    'local_gb': 250,
                    'vcpus_used': 2,
                    'memory_mb_used': 5 * 1024,
                    'local_gb_used': 125,
                    'hypervisor_type': 'xen',
                    'hypervisor_version': 3,
                    'hypervisor_hostname': 'hyper1',
                    'free_ram_mb': 5 * 1024,
                    'free_disk_gb': 125,
                    'current_workload': 2,
                    'running_vms': 2,
                    'cpu_info': 'cpu_info',
                    'disk_available_least': 100,
                    'state': 'up',
                    'status': 'enabled',
                },
                {
                    'id': self.hyper_id_2,
                    'service': {
                        'id': self.service_id_2,
                        'host': 'compute2',
                    },
                    'vcpus': 4,
                    'memory_mb': 10 * 1024,
                    'local_gb': 250,
                    'vcpus_used': 2,
                    'memory_mb_used': 5 * 1024,
                    'local_gb_used': 125,
                    'hypervisor_type': 'xen',
                    'hypervisor_version': 3,
                    'hypervisor_hostname': 'hyper2',
                    'free_ram_mb': 5 * 1024,
                    'free_disk_gb': 125,
                    'current_workload': 2,
                    'running_vms': 2,
                    'cpu_info': 'cpu_info',
                    'disk_available_least': 100,
                    'state': 'up',
                    'status': 'enabled',
                }
            ]
        }

        if api_version >= api_versions.APIVersion('2.88'):
            for hypervisor in get_os_hypervisors_detail['hypervisors']:
                self._transform_hypervisor_details(hypervisor)

        self.requests_mock.get(self.url('detail'),
                               json=get_os_hypervisors_detail,
                               headers=self.headers)

        get_os_hypervisors_stats = {
            'hypervisor_statistics': {
                'count': 2,
                'vcpus': 8,
                'memory_mb': 20 * 1024,
                'local_gb': 500,
                'vcpus_used': 4,
                'memory_mb_used': 10 * 1024,
                'local_gb_used': 250,
                'free_ram_mb': 10 * 1024,
                'free_disk_gb': 250,
                'current_workload': 4,
                'running_vms': 4,
                'disk_available_least': 200,
            }
        }

        self.requests_mock.get(self.url('statistics'),
                               json=get_os_hypervisors_stats,
                               headers=self.headers)

        get_os_hypervisors_search = {
            'hypervisors': [
                {
                    'id': self.hyper_id_1,
                    'hypervisor_hostname': 'hyper1',
                    'state': 'up',
                    'status': 'enabled',
                },
                {
                    'id': self.hyper_id_2,
                    'hypervisor_hostname': 'hyper2',
                    'state': 'up',
                    'status': 'enabled',
                },
            ]
        }

        if api_version >= api_versions.APIVersion('2.53'):
            url = self.url(hypervisor_hostname_pattern='hyper')
        else:
            url = self.url('hyper', 'search')
        self.requests_mock.get(url,
                               json=get_os_hypervisors_search,
                               headers=self.headers)

        if api_version >= api_versions.APIVersion('2.53'):
            get_os_hypervisors_search_u_v2_53 = {
                'error_name': 'BadRequest',
                'message': 'Invalid input for query parameters '
                           'hypervisor_hostname_pattern.',
                'code': 400}
            # hypervisor_hostname_pattern is encoded in the url method
            url = self.url(hypervisor_hostname_pattern='\\u5de5\\u4f5c')
            self.requests_mock.get(url,
                                   json=get_os_hypervisors_search_u_v2_53,
                                   headers=self.headers, status_code=400)
        else:
            get_os_hypervisors_search_unicode = {
                'error_name': 'NotFound',
                'message': "No hypervisor matching "
                           "'\\u5de5\\u4f5c' could be found.",
                'code': 404
            }
            hypervisor_hostname_pattern = parse.quote(encodeutils.safe_encode(
                '\\u5de5\\u4f5c'))
            url = self.url(hypervisor_hostname_pattern, 'search')
            self.requests_mock.get(url,
                                   json=get_os_hypervisors_search_unicode,
                                   headers=self.headers, status_code=404)

        get_hyper_server = {
            'hypervisors': [
                {
                    'id': self.hyper_id_1,
                    'hypervisor_hostname': 'hyper1',
                    'state': 'up',
                    'status': 'enabled',
                    'servers': [
                        {'name': 'inst1', 'uuid': 'uuid1'},
                        {'name': 'inst2', 'uuid': 'uuid2'}
                    ]
                },
                {
                    'id': self.hyper_id_2,
                    'hypervisor_hostname': 'hyper2',
                    'state': 'up',
                    'status': 'enabled',
                    'servers': [
                        {'name': 'inst3', 'uuid': 'uuid3'},
                        {'name': 'inst4', 'uuid': 'uuid4'}
                    ]
                }
            ]
        }

        if api_version >= api_versions.APIVersion('2.53'):
            url = self.url(hypervisor_hostname_pattern='hyper',
                           with_servers=True)
        else:
            url = self.url('hyper', 'servers')
        self.requests_mock.get(url,
                               json=get_hyper_server,
                               headers=self.headers)

        get_os_hypervisors_hyper1 = {
            'hypervisor': {
                'id': self.hyper_id_1,
                'service': {'id': self.service_id_1, 'host': 'compute1'},
                'vcpus': 4,
                'memory_mb': 10 * 1024,
                'local_gb': 250,
                'vcpus_used': 2,
                'memory_mb_used': 5 * 1024,
                'local_gb_used': 125,
                'hypervisor_type': 'xen',
                'hypervisor_version': 3,
                'hypervisor_hostname': 'hyper1',
                'free_ram_mb': 5 * 1024,
                'free_disk_gb': 125,
                'current_workload': 2,
                'running_vms': 2,
                'cpu_info': 'cpu_info',
                'disk_available_least': 100,
                'state': 'up',
                'status': 'enabled',
            }
        }

        if api_version >= api_versions.APIVersion('2.88'):
            self._transform_hypervisor_details(
                get_os_hypervisors_hyper1['hypervisor'])

        self.requests_mock.get(self.url(self.hyper_id_1),
                               json=get_os_hypervisors_hyper1,
                               headers=self.headers)

        get_os_hypervisors_uptime = {
            'hypervisor': {
                'id': self.hyper_id_1,
                'hypervisor_hostname': 'hyper1',
                'uptime': 'fake uptime',
                'state': 'up',
                'status': 'enabled',
            }
        }

        self.requests_mock.get(self.url(self.hyper_id_1, 'uptime'),
                               json=get_os_hypervisors_uptime,
                               headers=self.headers)


class V253(V1):
    """Fixture data for the os-hypervisors 2.53 API."""
    api_version = '2.53'
    hyper_id_1 = 'd480b1b6-2255-43c2-b2c2-d60d42c2c074'
    hyper_id_2 = '43a8214d-f36a-4fc0-a25c-3cf35c17522d'
    service_id_1 = 'a87743ff-9c29-42ff-805d-2444659b5fc0'
    service_id_2 = '0486ab8b-1cfc-4ccb-9d94-9f22ec8bbd6b'


class V288(V253):
    """Fixture data for the os-hypervisors 2.88 API."""
    api_version = '2.88'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

from tempest.lib import exceptions

from novaclient.tests.functional import base
from novaclient.tests.functional.v2 import fake_crypto


class TestKeypairsNovaClient(base.ClientTestBase):
    """Keypairs functional tests."""

    COMPUTE_API_VERSION = "2.1"

    def _serialize_kwargs(self, kwargs):
        kwargs_pairs = ['--%(key)s %(val)s' % {'key': key.replace('_', '-'),
                                               'val': val}
                        for key, val in kwargs.items()]
        return " ".join(kwargs_pairs)

    def _create_keypair(self, **kwargs):
        key_name = self._raw_create_keypair(**kwargs)
        self.addCleanup(self.nova, 'keypair-delete %s' % key_name)
        return key_name

    def _raw_create_keypair(self, **kwargs):
        key_name = self.name_generate()
        kwargs_str = self._serialize_kwargs(kwargs)
        self.nova('keypair-add %s %s' % (kwargs_str, key_name))
        return key_name

    def _show_keypair(self, key_name):
        return self.nova('keypair-show %s' % key_name)

    def _list_keypairs(self):
        return self.nova('keypair-list')

    def _delete_keypair(self, key_name):
        self.nova('keypair-delete %s' % key_name)

    def _create_public_key_file(self, public_key):
        pubfile = tempfile.mkstemp()[1]
        with open(pubfile, 'w') as f:
            f.write(public_key)
        return pubfile

    def test_create_keypair(self):
        key_name = self._create_keypair()
        keypair = self._show_keypair(key_name)
        self.assertIn(key_name, keypair)

        return keypair

    def _test_import_keypair(self, fingerprint, **create_kwargs):
        key_name = self._create_keypair(**create_kwargs)
        keypair = self._show_keypair(key_name)
        self.assertIn(key_name, keypair)
        self.assertIn(fingerprint, keypair)

        return keypair

    def test_import_keypair(self):
        pub_key, fingerprint = fake_crypto.get_ssh_pub_key_and_fingerprint()
        pub_key_file = self._create_public_key_file(pub_key)
        self._test_import_keypair(fingerprint, pub_key=pub_key_file)

    def test_list_keypair(self):
        key_name = self._create_keypair()
        keypairs = self._list_keypairs()
        self.assertIn(key_name, keypairs)

    def test_delete_keypair(self):
        key_name = self._raw_create_keypair()
        keypair = self._show_keypair(key_name)
        self.assertIsNotNone(keypair)

        self._delete_keypair(key_name)

        # keypair-show should fail if no keypair with given name is found.
        self.assertRaises(exceptions.CommandFailed,
                          self._show_keypair, key_name)



# -*- coding: utf-8 -*-
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient.tests.functional import base


class TestAggregatesNovaClient(base.ClientTestBase):
    COMPUTE_API_VERSION = '2.1'

    def setUp(self):
        super(TestAggregatesNovaClient, self).setUp()
        self.agg1 = self.name_generate()
        self.agg2 = self.name_generate()
        self.addCleanup(self._clean_aggregates)

    def _clean_aggregates(self):
        for a in (self.agg1, self.agg2):
            try:
                self.nova('aggregate-delete', params=a)
            except Exception:
                pass

    def test_aggregate_update_name(self):
        self.nova('aggregate-create', params=self.agg1)
        self.nova('aggregate-update',
                  params='--name=%s %s' % (self.agg2, self.agg1))
        output = self.nova('aggregate-show', params=self.agg2)
        self.assertIn(self.agg2, output)
        self.nova('aggregate-delete', params=self.agg2)

    def test_aggregate_update_az(self):
        self.nova('aggregate-create', params=self.agg2)
        self.nova('aggregate-update',
                  params='--availability-zone=myaz %s' % self.agg2)
        output = self.nova('aggregate-show', params=self.agg2)
        self.assertIn('myaz', output)
        self.nova('aggregate-delete', params=self.agg2)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes
from novaclient.v2 import migrations


class MigrationsTest(utils.TestCase):
    def setUp(self):
        super(MigrationsTest, self).setUp()
        self.cs = fakes.FakeClient(api_versions.APIVersion("2.1"))

    def test_list_migrations(self):
        ml = self.cs.migrations.list()
        self.assert_request_id(ml, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET', '/os-migrations')
        for m in ml:
            self.assertIsInstance(m, migrations.Migration)
            self.assertRaises(AttributeError, getattr, m, "migration_type")
            self.assertRaises(AttributeError, getattr, m, "uuid")

    def test_list_migrations_with_filters(self):
        ml = self.cs.migrations.list('host1', 'finished')
        self.assert_request_id(ml, fakes.FAKE_REQUEST_ID_LIST)

        self.cs.assert_called(
            'GET',
            '/os-migrations?host=host1&status=finished')
        for m in ml:
            self.assertIsInstance(m, migrations.Migration)

    def test_list_migrations_with_instance_uuid_filter(self):
        ml = self.cs.migrations.list('host1', 'finished', 'instance_id_456')
        self.assert_request_id(ml, fakes.FAKE_REQUEST_ID_LIST)

        self.cs.assert_called(
            'GET',
            ('/os-migrations?host=host1&'
             'instance_uuid=instance_id_456&status=finished'))
        self.assertEqual(1, len(ml))
        self.assertEqual('instance_id_456', ml[0].instance_uuid)


class MigrationsV223Test(MigrationsTest):
    def setUp(self):
        super(MigrationsV223Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.23")

    def test_list_migrations(self):
        ml = self.cs.migrations.list()
        self.assert_request_id(ml, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET', '/os-migrations')
        for m in ml:
            self.assertIsInstance(m, migrations.Migration)
            self.assertEqual(m.migration_type, 'live-migration')
            self.assertRaises(AttributeError, getattr, m, "uuid")


class MigrationsV259Test(MigrationsV223Test):
    def setUp(self):
        super(MigrationsV259Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.59")

    def test_list_migrations(self):
        ml = self.cs.migrations.list()
        self.assert_request_id(ml, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET', '/os-migrations')
        for m in ml:
            self.assertIsInstance(m, migrations.Migration)
            self.assertEqual(m.migration_type, 'live-migration')
            self.assertTrue(hasattr(m, 'uuid'))

    def test_list_migrations_with_limit_marker_params(self):
        marker = '12140183-c814-4ddf-8453-6df43028aa67'
        params = {'limit': 10,
                  'marker': marker,
                  'changes_since': '2012-02-29T06:23:22'}

        ms = self.cs.migrations.list(**params)
        self.assert_request_id(ms, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET',
                              '/os-migrations?'
                              'changes-since=%s&limit=10&marker=%s'
                              % ('2012-02-29T06%3A23%3A22', marker))
        for m in ms:
            self.assertIsInstance(m, migrations.Migration)


class MigrationsV266Test(MigrationsV259Test):
    def setUp(self):
        super(MigrationsV266Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.66")

    def test_list_migrations_with_changes_before(self):
        params = {'changes_before': '2012-02-29T06:23:22'}
        ms = self.cs.migrations.list(**params)
        self.assert_request_id(ms, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET',
                              '/os-migrations?'
                              'changes-before=%s' %
                              '2012-02-29T06%3A23%3A22')
        for m in ms:
            self.assertIsInstance(m, migrations.Migration)


class MigrationsV280Test(MigrationsV266Test):
    def setUp(self):
        super(MigrationsV280Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.80")

    def test_list_migrations_with_user_id(self):
        user_id = '13cc0930d27c4be0acc14d7c47a3e1f7'
        params = {'user_id': user_id}
        ms = self.cs.migrations.list(**params)
        self.assert_request_id(ms, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET', '/os-migrations?user_id=%s' % user_id)
        for m in ms:
            self.assertIsInstance(m, migrations.Migration)

    def test_list_migrations_with_project_id(self):
        project_id = 'b59c18e5fa284fd384987c5cb25a1853'
        params = {'project_id': project_id}
        ms = self.cs.migrations.list(**params)
        self.assert_request_id(ms, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET', '/os-migrations?project_id=%s'
                              % project_id)
        for m in ms:
            self.assertIsInstance(m, migrations.Migration)

    def test_list_migrations_with_user_and_project_id(self):
        user_id = '13cc0930d27c4be0acc14d7c47a3e1f7'
        project_id = 'b59c18e5fa284fd384987c5cb25a1853'
        params = {'user_id': user_id, 'project_id': project_id}
        ms = self.cs.migrations.list(**params)
        self.assert_request_id(ms, fakes.FAKE_REQUEST_ID_LIST)
        self.cs.assert_called('GET',
                              '/os-migrations?project_id=%s&user_id=%s'
                              % (project_id, user_id))
        for m in ms:
            self.assertIsInstance(m, migrations.Migration)

    def test_list_migrations_with_user_id_pre_v280(self):
        self.cs.api_version = api_versions.APIVersion('2.79')
        user_id = '13cc0930d27c4be0acc14d7c47a3e1f7'
        ex = self.assertRaises(TypeError,
                               self.cs.migrations.list,
                               user_id=user_id)
        self.assertIn("unexpected keyword argument 'user_id'", str(ex))

    def test_list_migrations_with_project_id_pre_v280(self):
        self.cs.api_version = api_versions.APIVersion('2.79')
        project_id = '23cc0930d27c4be0acc14d7c47a3e1f7'
        ex = self.assertRaises(TypeError,
                               self.cs.migrations.list,
                               project_id=project_id)
        self.assertIn("unexpected keyword argument 'project_id'", str(ex))



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient import base


class QuotaSet(base.Resource):

    def update(self, *args, **kwargs):
        return self.manager.update(self.id, *args, **kwargs)


class QuotaSetManager(base.Manager):
    resource_class = QuotaSet

    def get(self, tenant_id, user_id=None, detail=False):
        url = '/os-quota-sets/%(tenant_id)s'
        if detail:
            url += '/detail'

        if user_id:
            params = {'tenant_id': tenant_id, 'user_id': user_id}
            url += '?user_id=%(user_id)s'
        else:
            params = {'tenant_id': tenant_id}

        return self._get(url % params, "quota_set")

    # NOTE(mriedem): Before 2.57 the resources you could update was just a
    # kwargs dict and not validated on the client-side, only on the API server
    # side.
    @api_versions.wraps("2.0", "2.56")
    def update(self, tenant_id, **kwargs):

        user_id = kwargs.pop('user_id', None)
        body = {'quota_set': kwargs}

        for key in list(body['quota_set']):
            if body['quota_set'][key] is None:
                body['quota_set'].pop(key)

        if user_id:
            url = '/os-quota-sets/%s?user_id=%s' % (tenant_id, user_id)
        else:
            url = '/os-quota-sets/%s' % tenant_id
        return self._update(url, body, 'quota_set')

    # NOTE(mriedem): 2.57 does strict validation of the resources you can
    # specify. 2.36 blocks network-related resources and 2.57 blocks
    # injected files related quotas.
    @api_versions.wraps("2.57")
    def update(self, tenant_id, user_id=None, force=False,
               instances=None, cores=None, ram=None,
               metadata_items=None, key_pairs=None, server_groups=None,
               server_group_members=None):

        resources = {}
        if force:
            resources['force'] = force
        if instances is not None:
            resources['instances'] = instances
        if cores is not None:
            resources['cores'] = cores
        if ram is not None:
            resources['ram'] = ram
        if metadata_items is not None:
            resources['metadata_items'] = metadata_items
        if key_pairs is not None:
            resources['key_pairs'] = key_pairs
        if server_groups is not None:
            resources['server_groups'] = server_groups
        if server_group_members is not None:
            resources['server_group_members'] = server_group_members
        body = {'quota_set': resources}

        if user_id:
            url = '/os-quota-sets/%s?user_id=%s' % (tenant_id, user_id)
        else:
            url = '/os-quota-sets/%s' % tenant_id
        return self._update(url, body, 'quota_set')

    def defaults(self, tenant_id):
        return self._get('/os-quota-sets/%s/defaults' % tenant_id,
                         'quota_set')

    def delete(self, tenant_id, user_id=None):
        """
        Delete quota for a tenant or for a user.

        :param tenant_id: A tenant for which quota is to be deleted
        :param user_id: A user for which quota is to be deleted
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        if user_id:
            url = '/os-quota-sets/%s?user_id=%s' % (tenant_id, user_id)
        else:
            url = '/os-quota-sets/%s' % tenant_id
        return self._delete(url)



# Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import exceptions
from novaclient.tests.unit import utils as test_utils


class ExceptionsTestCase(test_utils.TestCase):

    def _test_from_response(self, body, expected_message):
        data = {
            'status_code': 404,
            'headers': {
                'content-type': 'application/json',
                'x-openstack-request-id': (
                    'req-d9df03b0-4150-4b53-8157-7560ccf39f75'),
            }
        }
        response = test_utils.TestResponse(data)
        fake_url = 'http://localhost:8774/v2.1/fake/flavors/test'
        error = exceptions.from_response(response, body, fake_url, 'GET')
        self.assertIsInstance(error, exceptions.NotFound)
        self.assertEqual(expected_message, error.message)

    def test_from_response_webob_pre_1_6_0(self):
        # Tests error responses before webob 1.6.0 where the error details
        # are nested in the response body.
        message = "Flavor test could not be found."
        self._test_from_response(
            {"itemNotFound": {"message": message, "code": 404}},
            message)

    def test_from_response_webob_post_1_6_0(self):
        # Tests error responses from webob 1.6.0 where the error details
        # are in the response body.
        message = "Flavor test could not be found."
        self._test_from_response({"message": message, "code": 404}, message)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import io
import re
import sys
from unittest import mock

import ddt
import fixtures
from keystoneauth1 import fixture
import requests_mock
from testtools import matchers

from novaclient import api_versions
import novaclient.client
from novaclient import exceptions
import novaclient.shell
from novaclient.tests.unit import fake_actions_module
from novaclient.tests.unit import utils

FAKE_ENV = {'OS_USERNAME': 'username',
            'OS_PASSWORD': 'password',
            'OS_TENANT_NAME': 'tenant_name',
            'OS_AUTH_URL': 'http://no.where/v3',
            'OS_COMPUTE_API_VERSION': '2',
            'OS_PROJECT_DOMAIN_ID': 'default',
            'OS_PROJECT_DOMAIN_NAME': 'default',
            'OS_USER_DOMAIN_ID': 'default',
            'OS_USER_DOMAIN_NAME': 'default'}

FAKE_ENV2 = {'OS_USER_ID': 'user_id',
             'OS_PASSWORD': 'password',
             'OS_TENANT_ID': 'tenant_id',
             'OS_AUTH_URL': 'http://no.where/v3',
             'OS_COMPUTE_API_VERSION': '2'}

FAKE_ENV3 = {'OS_USER_ID': 'user_id',
             'OS_PASSWORD': 'password',
             'OS_TENANT_ID': 'tenant_id',
             'OS_AUTH_URL': 'http://no.where/v3',
             'NOVA_ENDPOINT_TYPE': 'novaURL',
             'OS_ENDPOINT_TYPE': 'osURL',
             'OS_COMPUTE_API_VERSION': '2'}

FAKE_ENV4 = {'OS_USER_ID': 'user_id',
             'OS_PASSWORD': 'password',
             'OS_TENANT_ID': 'tenant_id',
             'OS_AUTH_URL': 'http://no.where/v3',
             'NOVA_ENDPOINT_TYPE': 'internal',
             'OS_ENDPOINT_TYPE': 'osURL',
             'OS_COMPUTE_API_VERSION': '2'}

FAKE_ENV5 = {'OS_USERNAME': 'username',
             'OS_PASSWORD': 'password',
             'OS_TENANT_NAME': 'tenant_name',
             'OS_AUTH_URL': 'http://no.where/v3'}


def _create_ver_list(versions):
    return {'versions': {'values': versions}}


class DeprecatedActionTest(utils.TestCase):
    @mock.patch.object(argparse.Action, '__init__', return_value=None)
    def test_init_emptyhelp_nouse(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertIsNone(result.use)
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest', 'Deprecated',
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated', a=1, b=2, c=3)

    @mock.patch.object(novaclient.shell.argparse.Action, '__init__',
                       return_value=None)
    def test_init_emptyhelp_withuse(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', use='use this instead', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertEqual(result.use, 'use this instead')
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest',
                          'Deprecated; use this instead',
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated; use this instead',
            a=1, b=2, c=3)

    @mock.patch.object(argparse.Action, '__init__', return_value=None)
    def test_init_withhelp_nouse(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', help='some help', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertIsNone(result.use)
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest',
                          'some help (Deprecated)',
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help='some help (Deprecated)',
            a=1, b=2, c=3)

    @mock.patch.object(novaclient.shell.argparse.Action, '__init__',
                       return_value=None)
    def test_init_withhelp_withuse(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', help='some help',
            use='use this instead', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertEqual(result.use, 'use this instead')
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest',
                          'some help (Deprecated; use this instead)',
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest',
            help='some help (Deprecated; use this instead)',
            a=1, b=2, c=3)

    @mock.patch.object(argparse.Action, '__init__', return_value=None)
    def test_init_suppresshelp_nouse(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', help=argparse.SUPPRESS, a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertIsNone(result.use)
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest', argparse.SUPPRESS,
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help=argparse.SUPPRESS, a=1, b=2, c=3)

    @mock.patch.object(novaclient.shell.argparse.Action, '__init__',
                       return_value=None)
    def test_init_suppresshelp_withuse(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', help=argparse.SUPPRESS,
            use='use this instead', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertEqual(result.use, 'use this instead')
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest', argparse.SUPPRESS,
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help=argparse.SUPPRESS, a=1, b=2, c=3)

    @mock.patch.object(argparse.Action, '__init__', return_value=None)
    def test_init_action_nothing(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', real_action='nothing', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertIsNone(result.use)
        self.assertIs(result.real_action_args, False)
        self.assertIsNone(result.real_action)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated', a=1, b=2, c=3)

    @mock.patch.object(argparse.Action, '__init__', return_value=None)
    def test_init_action_string(self, mock_init):
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', real_action='store', a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertIsNone(result.use)
        self.assertEqual(result.real_action_args,
                         ('option_strings', 'dest', 'Deprecated',
                          {'a': 1, 'b': 2, 'c': 3}))
        self.assertEqual(result.real_action, 'store')
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated', a=1, b=2, c=3)

    @mock.patch.object(argparse.Action, '__init__', return_value=None)
    def test_init_action_other(self, mock_init):
        action = mock.Mock()
        result = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', real_action=action, a=1, b=2, c=3)

        self.assertEqual(result.emitted, set())
        self.assertIsNone(result.use)
        self.assertIs(result.real_action_args, False)
        self.assertEqual(result.real_action, action.return_value)
        mock_init.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated', a=1, b=2, c=3)
        action.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated', a=1, b=2, c=3)

    @mock.patch.object(sys, 'stderr', io.StringIO())
    def test_get_action_nolookup(self):
        action_class = mock.Mock()
        parser = mock.Mock(**{
            '_registry_get.return_value': action_class,
        })
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', real_action='nothing', const=1)
        obj.real_action = 'action'

        result = obj._get_action(parser)

        self.assertEqual(result, 'action')
        self.assertEqual(obj.real_action, 'action')
        self.assertFalse(parser._registry_get.called)
        self.assertFalse(action_class.called)
        self.assertEqual(sys.stderr.getvalue(), '')

    @mock.patch.object(sys, 'stderr', io.StringIO())
    def test_get_action_lookup_noresult(self):
        parser = mock.Mock(**{
            '_registry_get.return_value': None,
        })
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', real_action='store', const=1)

        result = obj._get_action(parser)

        self.assertIsNone(result)
        self.assertIsNone(obj.real_action)
        parser._registry_get.assert_called_once_with(
            'action', 'store')
        self.assertEqual(sys.stderr.getvalue(),
                         'WARNING: Programming error: Unknown real action '
                         '"store"\n')

    @mock.patch.object(sys, 'stderr', io.StringIO())
    def test_get_action_lookup_withresult(self):
        action_class = mock.Mock()
        parser = mock.Mock(**{
            '_registry_get.return_value': action_class,
        })
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', real_action='store', const=1)

        result = obj._get_action(parser)

        self.assertEqual(result, action_class.return_value)
        self.assertEqual(obj.real_action, action_class.return_value)
        parser._registry_get.assert_called_once_with(
            'action', 'store')
        action_class.assert_called_once_with(
            'option_strings', 'dest', help='Deprecated', const=1)
        self.assertEqual(sys.stderr.getvalue(), '')

    @mock.patch.object(sys, 'stderr', io.StringIO())
    @mock.patch.object(novaclient.shell.DeprecatedAction, '_get_action')
    def test_call_unemitted_nouse(self, mock_get_action):
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest')

        obj('parser', 'namespace', 'values', 'option_string')

        self.assertEqual(obj.emitted, set(['option_string']))
        mock_get_action.assert_called_once_with('parser')
        mock_get_action.return_value.assert_called_once_with(
            'parser', 'namespace', 'values', 'option_string')
        self.assertEqual(sys.stderr.getvalue(),
                         'WARNING: Option "option_string" is deprecated\n')

    @mock.patch.object(sys, 'stderr', io.StringIO())
    @mock.patch.object(novaclient.shell.DeprecatedAction, '_get_action')
    def test_call_unemitted_withuse(self, mock_get_action):
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', use='use this instead')

        obj('parser', 'namespace', 'values', 'option_string')

        self.assertEqual(obj.emitted, set(['option_string']))
        mock_get_action.assert_called_once_with('parser')
        mock_get_action.return_value.assert_called_once_with(
            'parser', 'namespace', 'values', 'option_string')
        self.assertEqual(sys.stderr.getvalue(),
                         'WARNING: Option "option_string" is deprecated; '
                         'use this instead\n')

    @mock.patch.object(sys, 'stderr', io.StringIO())
    @mock.patch.object(novaclient.shell.DeprecatedAction, '_get_action')
    def test_call_emitted_nouse(self, mock_get_action):
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest')
        obj.emitted.add('option_string')

        obj('parser', 'namespace', 'values', 'option_string')

        self.assertEqual(obj.emitted, set(['option_string']))
        mock_get_action.assert_called_once_with('parser')
        mock_get_action.return_value.assert_called_once_with(
            'parser', 'namespace', 'values', 'option_string')
        self.assertEqual(sys.stderr.getvalue(), '')

    @mock.patch.object(sys, 'stderr', io.StringIO())
    @mock.patch.object(novaclient.shell.DeprecatedAction, '_get_action')
    def test_call_emitted_withuse(self, mock_get_action):
        obj = novaclient.shell.DeprecatedAction(
            'option_strings', 'dest', use='use this instead')
        obj.emitted.add('option_string')

        obj('parser', 'namespace', 'values', 'option_string')

        self.assertEqual(obj.emitted, set(['option_string']))
        mock_get_action.assert_called_once_with('parser')
        mock_get_action.return_value.assert_called_once_with(
            'parser', 'namespace', 'values', 'option_string')
        self.assertEqual(sys.stderr.getvalue(), '')


class ParserTest(utils.TestCase):

    def setUp(self):
        super(ParserTest, self).setUp()
        self.parser = novaclient.shell.NovaClientArgumentParser()

    def test_ambiguous_option(self):
        self.parser.add_argument('--tic')
        self.parser.add_argument('--tac')

        try:
            self.parser.parse_args(['--t'])
        except SystemExit as err:
            self.assertEqual(2, err.code)
        else:
            self.fail('SystemExit not raised')

    def test_not_really_ambiguous_option(self):
        # current/deprecated forms of the same option
        self.parser.add_argument('--tic-tac', action="store_true")
        self.parser.add_argument('--tic_tac', action="store_true")
        args = self.parser.parse_args(['--tic'])
        self.assertTrue(args.tic_tac)


@ddt.ddt
class ShellTest(utils.TestCase):

    _msg_no_tenant_project = ("You must provide a project name or project"
                              " ID via --os-project-name, --os-project-id,"
                              " env[OS_PROJECT_ID] or env[OS_PROJECT_NAME]."
                              " You may use os-project and os-tenant"
                              " interchangeably.")

    def make_env(self, exclude=None, fake_env=FAKE_ENV):
        env = dict((k, v) for k, v in fake_env.items() if k != exclude)
        self.useFixture(fixtures.MonkeyPatch('os.environ', env))

    def setUp(self):
        super(ShellTest, self).setUp()
        self.mock_client = mock.MagicMock()
        self.mock_client.return_value.api_version = novaclient.API_MIN_VERSION
        self.useFixture(fixtures.MonkeyPatch('novaclient.client.Client',
                                             self.mock_client))
        self.nc_util = mock.patch('novaclient.utils.isunauthenticated').start()
        self.nc_util.return_value = False
        self.mock_server_version_range = mock.patch(
            'novaclient.api_versions._get_server_version_range').start()
        self.mock_server_version_range.return_value = (
            novaclient.API_MIN_VERSION,
            novaclient.API_MIN_VERSION)
        self.orig_max_ver = novaclient.API_MAX_VERSION
        self.orig_min_ver = novaclient.API_MIN_VERSION
        self.addCleanup(self._clear_fake_version)
        self.addCleanup(mock.patch.stopall)

    def _clear_fake_version(self):
        novaclient.API_MAX_VERSION = self.orig_max_ver
        novaclient.API_MIN_VERSION = self.orig_min_ver

    def shell(self, argstr, exitcodes=(0,)):
        orig = sys.stdout
        orig_stderr = sys.stderr
        try:
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()
            _shell = novaclient.shell.OpenStackComputeShell()
            _shell.main(argstr.split())
        except SystemExit:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.assertIn(exc_value.code, exitcodes)
        finally:
            stdout = sys.stdout.getvalue()
            sys.stdout.close()
            sys.stdout = orig
            stderr = sys.stderr.getvalue()
            sys.stderr.close()
            sys.stderr = orig_stderr
        return (stdout, stderr)

    def register_keystone_discovery_fixture(self, mreq):
        v3_url = "http://no.where/v3"
        v3_version = fixture.V3Discovery(v3_url)
        mreq.register_uri(
            'GET', v3_url, json=_create_ver_list([v3_version]),
            status_code=200)

    def test_help_unknown_command(self):
        self.assertRaises(exceptions.CommandError, self.shell, 'help foofoo')

    def test_invalid_timeout(self):
        for f in [0, -1, -10]:
            cmd_text = '--timeout %s' % (f)
            stdout, stderr = self.shell(cmd_text, exitcodes=[0, 2])
            required = [
                'argument --timeout: %s must be greater than 0' % (f),
            ]
            for r in required:
                self.assertIn(r, stderr)

    def _test_help(self, command, required=None):
        if required is None:
            required = [
                '.*?^usage: ',
                '.*?^\\s+set-password\\s+Change the admin password',
                '.*?^See "nova help COMMAND" for help on a specific command',
            ]
        stdout, stderr = self.shell(command)
        for r in required:
            self.assertThat((stdout + stderr),
                            matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))

    def test_help(self):
        self._test_help('help')

    def test_help_option(self):
        self._test_help('--help')
        self._test_help('-h')

    def test_help_no_options(self):
        self._test_help('')

    def test_help_no_subcommand(self):
        self._test_help('--os-compute-api-version 2.87')

    def test_help_on_subcommand(self):
        required = [
            '.*?^usage: nova set-password',
            '.*?^Change the admin password',
            '.*?^Positional arguments:',
        ]
        self._test_help('help set-password', required=required)

    def test_bash_completion(self):
        stdout, stderr = self.shell('bash-completion')
        # just check we have some output
        required = [
            '.*--matching',
            '.*--wrap',
            '.*help',
            '.*server-group-delete',
            '.*--image-with']
        for r in required:
            self.assertThat((stdout + stderr),
                            matchers.MatchesRegex(r, re.DOTALL | re.MULTILINE))

    def test_no_username(self):
        required = ('You must provide a user name/id (via --os-username, '
                    '--os-user-id, env[OS_USERNAME] or env[OS_USER_ID]) or '
                    'an auth token (via --os-token).')
        self.make_env(exclude='OS_USERNAME')
        try:
            self.shell('list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_user_id(self):
        required = ('You must provide a user name/id (via --os-username, '
                    '--os-user-id, env[OS_USERNAME] or env[OS_USER_ID]) or '
                    'an auth token (via --os-token).')
        self.make_env(exclude='OS_USER_ID', fake_env=FAKE_ENV2)
        try:
            self.shell('list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_tenant_name(self):
        required = self._msg_no_tenant_project
        self.make_env(exclude='OS_TENANT_NAME')
        try:
            self.shell('list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_tenant_id(self):
        required = self._msg_no_tenant_project
        self.make_env(exclude='OS_TENANT_ID', fake_env=FAKE_ENV2)
        try:
            self.shell('list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args[0])
        else:
            self.fail('CommandError not raised')

    def test_no_auth_url(self):
        required = ('You must provide an auth url'
                    ' via either --os-auth-url or env[OS_AUTH_URL].',)
        self.make_env(exclude='OS_AUTH_URL')
        try:
            self.shell('list')
        except exceptions.CommandError as message:
            self.assertEqual(required, message.args)
        else:
            self.fail('CommandError not raised')

    @ddt.data(
        (None, 'project_domain_id', FAKE_ENV['OS_PROJECT_DOMAIN_ID']),
        ('OS_PROJECT_DOMAIN_ID', 'project_domain_id', ''),
        (None, 'project_domain_name', FAKE_ENV['OS_PROJECT_DOMAIN_NAME']),
        ('OS_PROJECT_DOMAIN_NAME', 'project_domain_name', ''),
        (None, 'user_domain_id', FAKE_ENV['OS_USER_DOMAIN_ID']),
        ('OS_USER_DOMAIN_ID', 'user_domain_id', ''),
        (None, 'user_domain_name', FAKE_ENV['OS_USER_DOMAIN_NAME']),
        ('OS_USER_DOMAIN_NAME', 'user_domain_name', '')
    )
    @ddt.unpack
    def test_basic_attributes(self, exclude, client_arg, env_var):
        self.make_env(exclude=exclude, fake_env=FAKE_ENV)
        self.shell('list')
        client_kwargs = self.mock_client.call_args_list[0][1]
        self.assertEqual(env_var, client_kwargs[client_arg])

    @requests_mock.Mocker()
    def test_nova_endpoint_type(self, m_requests):
        self.make_env(fake_env=FAKE_ENV3)
        self.register_keystone_discovery_fixture(m_requests)
        self.shell('list')
        client_kwargs = self.mock_client.call_args_list[0][1]
        self.assertEqual(client_kwargs['endpoint_type'], 'novaURL')

    @requests_mock.Mocker()
    def test_endpoint_type_like_other_clients(self, m_requests):
        self.make_env(fake_env=FAKE_ENV4)
        self.register_keystone_discovery_fixture(m_requests)
        self.shell('list')
        client_kwargs = self.mock_client.call_args_list[0][1]
        self.assertEqual(client_kwargs['endpoint_type'], 'internalURL')

    @requests_mock.Mocker()
    def test_os_endpoint_type(self, m_requests):
        self.make_env(exclude='NOVA_ENDPOINT_TYPE', fake_env=FAKE_ENV3)
        self.register_keystone_discovery_fixture(m_requests)
        self.shell('list')
        client_kwargs = self.mock_client.call_args_list[0][1]
        self.assertEqual(client_kwargs['endpoint_type'], 'osURL')

    def test_default_endpoint_type(self):
        self.make_env()
        self.shell('list')
        client_kwargs = self.mock_client.call_args_list[0][1]
        self.assertEqual(client_kwargs['endpoint_type'], 'publicURL')

    @mock.patch('sys.stdin', side_effect=mock.MagicMock)
    @mock.patch('getpass.getpass', return_value='password')
    @requests_mock.Mocker()
    def test_password(self, mock_getpass, mock_stdin, m_requests):
        mock_stdin.encoding = "utf-8"

        ex = '\n'.join([
            '+----+------+--------+------------+-------------+----------+',
            '| ID | Name | Status | Task State | Power State | Networks |',
            '+----+------+--------+------------+-------------+----------+',
            '+----+------+--------+------------+-------------+----------+',
            ''
        ])
        self.make_env(exclude='OS_PASSWORD')
        self.register_keystone_discovery_fixture(m_requests)
        stdout, stderr = self.shell('list')
        self.assertEqual((stdout + stderr), ex)

    def _test_service_type(self, version, service_type, mock_client):
        if version is None:
            cmd = 'list'
        else:
            cmd = ('--service-type %s --os-compute-api-version %s list' %
                   (service_type, version))
        self.make_env()
        self.shell(cmd)
        _client_args, client_kwargs = mock_client.call_args_list[0]
        self.assertEqual(service_type, client_kwargs['service_type'])

    def test_default_service_type(self):
        self._test_service_type(None, 'compute', self.mock_client)

    def test_v2_service_type(self):
        self._test_service_type('2', 'compute', self.mock_client)

    def test_v_unknown_service_type(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          self._test_service_type,
                          'unknown', 'compute', self.mock_client)

    @mock.patch('sys.stdout', io.StringIO())
    @mock.patch('sys.stderr', io.StringIO())
    def test_main_noargs(self):
        # Ensure that main works with no command-line arguments
        try:
            novaclient.shell.main([])
        except SystemExit:
            self.fail('Unexpected SystemExit')

        # We expect the normal usage as a result
        self.assertIn(
            'Command-line interface to the OpenStack Nova API',
            sys.stdout.getvalue(),
        )
        # We also expect to see the deprecation warning
        self.assertIn(
            'nova CLI is deprecated and will be removed in a future release',
            sys.stderr.getvalue(),
        )

    @mock.patch.object(novaclient.shell.OpenStackComputeShell, 'main')
    def test_main_keyboard_interrupt(self, mock_compute_shell):
        # Ensure that exit code is 130 for KeyboardInterrupt
        mock_compute_shell.side_effect = KeyboardInterrupt()
        try:
            novaclient.shell.main([])
        except SystemExit as ex:
            self.assertEqual(ex.code, 130)

    @mock.patch.object(novaclient.shell.OpenStackComputeShell, 'times')
    @requests_mock.Mocker()
    def test_timing(self, m_times, m_requests):
        m_times.append.side_effect = RuntimeError('Boom!')
        self.make_env()
        self.register_keystone_discovery_fixture(m_requests)
        self.shell('list')
        exc = self.assertRaises(RuntimeError, self.shell, '--timings list')
        self.assertEqual('Boom!', str(exc))

    @requests_mock.Mocker()
    def test_osprofiler(self, m_requests):
        self.make_env()

        def client(*args, **kwargs):
            self.assertEqual('swordfish', kwargs['profile'])
        with mock.patch('novaclient.client.Client', client):
            # we are only interested in the fact Client is initialized properly
            self.shell('list --profile swordfish', (0, 2))

    @requests_mock.Mocker()
    def test_osprofiler_not_installed(self, m_requests):
        self.make_env()

        # NOTE(rpodolyaka): osprofiler is in test-requirements, so we have to
        # simulate its absence here
        with mock.patch('novaclient.shell.osprofiler_profiler', None):
            _, stderr = self.shell('list --profile swordfish', (0, 2))
            self.assertIn('unrecognized arguments: --profile swordfish',
                          stderr)

    def test_microversion_with_default_behaviour(self):
        self.make_env(fake_env=FAKE_ENV5)
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion("2.1"), api_versions.APIVersion("2.3"))
        self.shell('list')
        client_args = self.mock_client.call_args_list[1][0]
        self.assertEqual(api_versions.APIVersion("2.3"), client_args[0])

    def test_microversion_with_default_behaviour_with_legacy_server(self):
        self.make_env(fake_env=FAKE_ENV5)
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion(), api_versions.APIVersion())
        self.shell('list')
        client_args = self.mock_client.call_args_list[1][0]
        self.assertEqual(api_versions.APIVersion("2.0"), client_args[0])

    def test_microversion_with_latest(self):
        self.make_env()
        novaclient.API_MAX_VERSION = api_versions.APIVersion('2.3')
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion("2.1"), api_versions.APIVersion("2.3"))
        self.shell('--os-compute-api-version 2.latest list')
        client_args = self.mock_client.call_args_list[1][0]
        self.assertEqual(api_versions.APIVersion("2.3"), client_args[0])

    def test_microversion_with_specified_version(self):
        self.make_env()
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion("2.10"), api_versions.APIVersion("2.100"))
        novaclient.API_MAX_VERSION = api_versions.APIVersion("2.100")
        novaclient.API_MIN_VERSION = api_versions.APIVersion("2.90")
        self.shell('--os-compute-api-version 2.99 list')
        client_args = self.mock_client.call_args_list[1][0]
        self.assertEqual(api_versions.APIVersion("2.99"), client_args[0])

    def test_microversion_with_specified_version_out_of_range(self):
        novaclient.API_MAX_VERSION = api_versions.APIVersion("2.100")
        novaclient.API_MIN_VERSION = api_versions.APIVersion("2.90")
        self.assertRaises(exceptions.CommandError,
                          self.shell, '--os-compute-api-version 2.199 list')

    def test_microversion_with_v2_and_v2_1_server(self):
        self.make_env()
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion('2.1'), api_versions.APIVersion('2.3'))
        novaclient.API_MAX_VERSION = api_versions.APIVersion("2.100")
        novaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")
        self.shell('--os-compute-api-version 2 list')
        client_args = self.mock_client.call_args_list[1][0]
        self.assertEqual(api_versions.APIVersion("2.0"), client_args[0])

    def test_microversion_with_v2_and_v2_server(self):
        self.make_env()
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion(), api_versions.APIVersion())
        novaclient.API_MAX_VERSION = api_versions.APIVersion("2.100")
        novaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")
        self.shell('--os-compute-api-version 2 list')
        client_args = self.mock_client.call_args_list[1][0]
        self.assertEqual(api_versions.APIVersion("2.0"), client_args[0])

    def test_microversion_with_v2_without_server_compatible(self):
        self.make_env()
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion('2.2'), api_versions.APIVersion('2.3'))
        novaclient.API_MAX_VERSION = api_versions.APIVersion("2.100")
        novaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")
        self.assertRaises(
            exceptions.UnsupportedVersion,
            self.shell, '--os-compute-api-version 2 list')

    def test_microversion_with_specific_version_without_microversions(self):
        self.make_env()
        self.mock_server_version_range.return_value = (
            api_versions.APIVersion(), api_versions.APIVersion())
        novaclient.API_MAX_VERSION = api_versions.APIVersion("2.100")
        novaclient.API_MIN_VERSION = api_versions.APIVersion("2.1")
        self.assertRaises(
            exceptions.UnsupportedVersion,
            self.shell,
            '--os-compute-api-version 2.3 list')

    @mock.patch.object(novaclient.shell.OpenStackComputeShell, 'main')
    def test_main_error_handling(self, mock_compute_shell):
        class MyException(Exception):
            pass
        with mock.patch('sys.stderr', io.StringIO()):
            mock_compute_shell.side_effect = MyException('message')
            self.assertRaises(SystemExit, novaclient.shell.main, [])
            err = sys.stderr.getvalue()
        # We expect to see the error propagated
        self.assertIn('ERROR (MyException): message\n', err)
        # We also expect to see the deprecation warning
        self.assertIn(
            'nova CLI is deprecated and will be removed in a future release',
            err,
        )


class TestLoadVersionedActions(utils.TestCase):

    def test_load_versioned_actions(self):
        # first load with API version 2.15, ensuring we use the 2.15 version of
        # the underlying function (which returns 1)
        parser = novaclient.shell.NovaClientArgumentParser()
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.15"), False)
        self.assertIn('fake-action', shell.subcommands.keys())
        self.assertEqual(
            1, shell.subcommands['fake-action'].get_default('func')())

        # now load with API version 2.25, ensuring we now use the
        # correspponding version of the underlying function (which now returns
        # 2)
        parser = novaclient.shell.NovaClientArgumentParser()
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.25"), False)
        self.assertIn('fake-action', shell.subcommands.keys())
        self.assertEqual(
            2, shell.subcommands['fake-action'].get_default('func')())

    def test_load_versioned_actions_not_in_version_range(self):
        parser = novaclient.shell.NovaClientArgumentParser()
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.10000"), False)
        self.assertNotIn('fake-action', shell.subcommands.keys())
        self.assertIn('fake-action2', shell.subcommands.keys())

    def test_load_versioned_actions_with_help(self):
        parser = novaclient.shell.NovaClientArgumentParser()
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.15"), True)
        self.assertIn('fake-action', shell.subcommands.keys())
        expected_desc = (" (Supported by API versions '%(start)s' - "
                         "'%(end)s')") % {'start': '2.10', 'end': '2.30'}
        self.assertEqual(expected_desc,
                         shell.subcommands['fake-action'].description)

    def test_load_versioned_actions_with_help_on_latest(self):
        parser = novaclient.shell.NovaClientArgumentParser()
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.latest"), True)
        self.assertIn('another-fake-action', shell.subcommands.keys())
        expected_desc = (" (Supported by API versions '%(start)s' - "
                         "'%(end)s')%(hint)s") % {
            'start': '2.0', 'end': '2.latest',
            'hint': novaclient.shell.HINT_HELP_MSG}
        self.assertEqual(expected_desc,
                         shell.subcommands['another-fake-action'].description)

    @mock.patch.object(novaclient.shell.NovaClientArgumentParser,
                       'add_argument')
    def test_load_versioned_actions_with_args(self, mock_add_arg):
        parser = novaclient.shell.NovaClientArgumentParser(add_help=False)
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.1"), False)
        self.assertIn('fake-action2', shell.subcommands.keys())
        mock_add_arg.assert_has_calls([
            mock.call('-h', '--help', action='help', help='==SUPPRESS=='),
            mock.call('--foo')])

    @mock.patch.object(novaclient.shell.NovaClientArgumentParser,
                       'add_argument')
    def test_load_versioned_actions_with_args2(self, mock_add_arg):
        parser = novaclient.shell.NovaClientArgumentParser(add_help=False)
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.4"), False)
        self.assertIn('fake-action2', shell.subcommands.keys())
        mock_add_arg.assert_has_calls([
            mock.call('-h', '--help', action='help', help='==SUPPRESS=='),
            mock.call('--bar')])

    @mock.patch.object(novaclient.shell.NovaClientArgumentParser,
                       'add_argument')
    def test_load_versioned_actions_with_args_not_in_version_range(
            self, mock_add_arg):
        parser = novaclient.shell.NovaClientArgumentParser(add_help=False)
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.10000"), False)
        self.assertIn('fake-action2', shell.subcommands.keys())
        mock_add_arg.assert_has_calls([
            mock.call('-h', '--help', action='help', help='==SUPPRESS==')])

    @mock.patch.object(novaclient.shell.NovaClientArgumentParser,
                       'add_argument')
    def test_load_versioned_actions_with_args_and_help(self, mock_add_arg):
        parser = novaclient.shell.NovaClientArgumentParser(add_help=False)
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.4"), True)
        mock_add_arg.assert_has_calls([
            mock.call('-h', '--help', action='help', help='==SUPPRESS=='),
            mock.call('--bar',
                      help=" (Supported by API versions '2.3' - '2.4')")])

    @mock.patch.object(novaclient.shell.NovaClientArgumentParser,
                       'add_argument')
    def test_load_actions_with_versioned_args(self, mock_add_arg):
        parser = novaclient.shell.NovaClientArgumentParser(add_help=False)
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.20"), False)
        self.assertIn(mock.call('--foo', help="first foo"),
                      mock_add_arg.call_args_list)
        self.assertNotIn(mock.call('--foo', help="second foo"),
                         mock_add_arg.call_args_list)

        mock_add_arg.reset_mock()

        parser = novaclient.shell.NovaClientArgumentParser(add_help=False)
        subparsers = parser.add_subparsers(metavar='<subcommand>')
        shell = novaclient.shell.OpenStackComputeShell()
        shell.subcommands = {}
        shell._find_actions(subparsers, fake_actions_module,
                            api_versions.APIVersion("2.21"), False)
        self.assertNotIn(mock.call('--foo', help="first foo"),
                         mock_add_arg.call_args_list)
        self.assertIn(mock.call('--foo', help="second foo"),
                      mock_add_arg.call_args_list)


class ShellTestKeystoneV3(ShellTest):
    def make_env(self, exclude=None, fake_env=FAKE_ENV):
        if 'OS_AUTH_URL' in fake_env:
            fake_env.update({'OS_AUTH_URL': 'http://no.where/v3'})
        env = dict((k, v) for k, v in fake_env.items() if k != exclude)
        self.useFixture(fixtures.MonkeyPatch('os.environ', env))

    def register_keystone_discovery_fixture(self, mreq):
        v3_url = "http://no.where/v3"
        v3_version = fixture.V3Discovery(v3_url)
        mreq.register_uri(
            'GET', v3_url, json=_create_ver_list([v3_version]),
            status_code=200)



#    Copyright 2015 Huawei Technology corp.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient.tests.functional.v2.legacy import test_server_groups


class TestServerGroupClientV213(test_server_groups.TestServerGroupClient):
    """Server groups v2.13 functional tests."""

    COMPUTE_API_VERSION = "2.13"
    expected_metadata = True
    expected_policy_rules = False

    def test_create_server_group(self):
        sg_id = self._create_sg("affinity")
        self.addCleanup(self.nova, 'server-group-delete %s' % sg_id)
        sg = self.nova('server-group-get %s' % sg_id)
        result = self._get_column_value_from_single_row_table(sg, "Id")
        self._get_column_value_from_single_row_table(
            sg, "User Id")
        self._get_column_value_from_single_row_table(
            sg, "Project Id")
        self.assertEqual(sg_id, result)
        self._get_column_value_from_single_row_table(sg, "Metadata")
        self.assertIn(
            'affinity',
            self._get_column_value_from_single_row_table(sg, 'Policies'))
        self.assertNotIn('Rules', sg)

    def test_list_server_groups(self):
        sg_id = self._create_sg("affinity")
        self.addCleanup(self.nova, 'server-group-delete %s' % sg_id)
        sg = self.nova("server-group-list")
        result = self._get_column_value_from_single_row_table(sg, "Id")
        self._get_column_value_from_single_row_table(
            sg, "User Id")
        self._get_column_value_from_single_row_table(
            sg, "Project Id")
        self.assertEqual(sg_id, result)
        if self.expected_metadata:
            self._get_column_value_from_single_row_table(sg, "Metadata")
        else:
            self.assertNotIn(sg, 'Metadata')
        if self.expected_policy_rules:
            self.assertEqual(
                'affinity',
                self._get_column_value_from_single_row_table(sg, "Policy"))
            self.assertEqual(
                '{}',
                self._get_column_value_from_single_row_table(sg, "Rules"))
        else:
            self.assertIn(
                'affinity',
                self._get_column_value_from_single_row_table(sg, 'Policies'))
            self.assertNotIn('Rules', sg)

    def test_get_server_group(self):
        sg_id = self._create_sg("affinity")
        self.addCleanup(self.nova, 'server-group-delete %s' % sg_id)
        sg = self.nova('server-group-get %s' % sg_id)
        result = self._get_column_value_from_single_row_table(sg, "Id")
        self._get_column_value_from_single_row_table(
            sg, "User Id")
        self._get_column_value_from_single_row_table(
            sg, "Project Id")
        self.assertEqual(sg_id, result)
        if self.expected_metadata:
            self._get_column_value_from_single_row_table(sg, "Metadata")
        else:
            self.assertNotIn(sg, 'Metadata')
        if self.expected_policy_rules:
            self.assertEqual(
                'affinity',
                self._get_column_value_from_single_row_table(sg, "Policy"))
            self.assertEqual(
                '{}',
                self._get_column_value_from_single_row_table(sg, "Rules"))
        else:
            self.assertIn(
                'affinity',
                self._get_column_value_from_single_row_table(sg, 'Policies'))
            self.assertNotIn('Rules', sg)


class TestServerGroupClientV264(TestServerGroupClientV213):
    """Server groups v2.64 functional tests."""

    COMPUTE_API_VERSION = "2.64"
    expected_metadata = False
    expected_policy_rules = True

    def test_create_server_group(self):
        output = self.nova('server-group-create complex-anti-affinity-group '
                           'anti-affinity --rule max_server_per_host=3')
        sg_id = self._get_column_value_from_single_row_table(output, "Id")
        self.addCleanup(self.nova, 'server-group-delete %s' % sg_id)
        sg = self.nova('server-group-get %s' % sg_id)
        result = self._get_column_value_from_single_row_table(sg, "Id")
        self.assertEqual(sg_id, result)
        self._get_column_value_from_single_row_table(
            sg, "User Id")
        self._get_column_value_from_single_row_table(
            sg, "Project Id")
        self.assertNotIn('Metadata', sg)
        self.assertEqual(
            'anti-affinity',
            self._get_column_value_from_single_row_table(sg, "Policy"))
        self.assertIn(
            'max_server_per_host',
            self._get_column_value_from_single_row_table(sg, "Rules"))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import logging
import os
import pkgutil
import re
import traceback
import warnings

from oslo_utils import strutils

import novaclient
from novaclient import exceptions
from novaclient.i18n import _

LOG = logging.getLogger(__name__)
if not LOG.handlers:
    LOG.addHandler(logging.StreamHandler())


LEGACY_HEADER_NAME = "X-OpenStack-Nova-API-Version"
HEADER_NAME = "OpenStack-API-Version"
SERVICE_TYPE = "compute"

_SUBSTITUTIONS = {}

_type_error_msg = _("'%(other)s' should be an instance of '%(cls)s'")


class APIVersion(object):
    """This class represents an API Version Request.

    This class provides convenience methods for manipulation
    and comparison of version numbers that we need to do to
    implement microversions.
    """

    def __init__(self, version_str=None):
        """Create an API version object.

        :param version_str: String representation of APIVersionRequest.
                            Correct format is 'X.Y', where 'X' and 'Y'
                            are int values. None value should be used
                            to create Null APIVersionRequest, which is
                            equal to 0.0
        """
        self.ver_major = 0
        self.ver_minor = 0

        if version_str is not None:
            match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0|latest)$", version_str)
            if match:
                self.ver_major = int(match.group(1))
                if match.group(2) == "latest":
                    # NOTE(andreykurilin): Infinity allows to easily determine
                    # latest version and doesn't require any additional checks
                    # in comparison methods.
                    self.ver_minor = float("inf")
                else:
                    self.ver_minor = int(match.group(2))
            else:
                msg = _("Invalid format of client version '%s'. "
                        "Expected format 'X.Y', where X is a major part and Y "
                        "is a minor part of version.") % version_str
                raise exceptions.UnsupportedVersion(msg)

    def __str__(self):
        """Debug/Logging representation of object."""
        if self.is_latest():
            return "Latest API Version Major: %s" % self.ver_major
        return ("API Version Major: %s, Minor: %s"
                % (self.ver_major, self.ver_minor))

    def __repr__(self):
        if self.is_null():
            return "<APIVersion: null>"
        else:
            return "<APIVersion: %s>" % self.get_string()

    def is_null(self):
        return self.ver_major == 0 and self.ver_minor == 0

    def is_latest(self):
        return self.ver_minor == float("inf")

    def __lt__(self, other):
        if not isinstance(other, APIVersion):
            raise TypeError(_type_error_msg % {"other": other,
                                               "cls": self.__class__})

        return ((self.ver_major, self.ver_minor) <
                (other.ver_major, other.ver_minor))

    def __eq__(self, other):
        if not isinstance(other, APIVersion):
            raise TypeError(_type_error_msg % {"other": other,
                                               "cls": self.__class__})

        return ((self.ver_major, self.ver_minor) ==
                (other.ver_major, other.ver_minor))

    def __gt__(self, other):
        if not isinstance(other, APIVersion):
            raise TypeError(_type_error_msg % {"other": other,
                                               "cls": self.__class__})

        return ((self.ver_major, self.ver_minor) >
                (other.ver_major, other.ver_minor))

    def __le__(self, other):
        return self < other or self == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self > other or self == other

    def matches(self, min_version, max_version):
        """Matches the version object.

        Returns whether the version object represents a version
        greater than or equal to the minimum version and less than
        or equal to the maximum version.

        :param min_version: Minimum acceptable version.
        :param max_version: Maximum acceptable version.
        :returns: boolean

        If min_version is null then there is no minimum limit.
        If max_version is null then there is no maximum limit.
        If self is null then raise ValueError
        """

        if self.is_null():
            raise ValueError(_("Null APIVersion doesn't support 'matches'."))
        if max_version.is_null() and min_version.is_null():
            return True
        elif max_version.is_null():
            return min_version <= self
        elif min_version.is_null():
            return self <= max_version
        else:
            return min_version <= self <= max_version

    def get_string(self):
        """Version string representation.

        Converts object to string representation which if used to create
        an APIVersion object results in the same version.
        """
        if self.is_null():
            raise ValueError(
                _("Null APIVersion cannot be converted to string."))
        elif self.is_latest():
            return "%s.%s" % (self.ver_major, "latest")
        return "%s.%s" % (self.ver_major, self.ver_minor)


class VersionedMethod(object):

    def __init__(self, name, start_version, end_version, func):
        """Versioning information for a single method

        :param name: Name of the method
        :param start_version: Minimum acceptable version
        :param end_version: Maximum acceptable_version
        :param func: Method to call

        Minimum and maximums are inclusive
        """
        self.name = name
        self.start_version = start_version
        self.end_version = end_version
        self.func = func

    def __str__(self):
        return ("Version Method %s: min: %s, max: %s"
                % (self.name, self.start_version, self.end_version))

    def __repr__(self):
        return "<VersionedMethod %s>" % self.name


def get_available_major_versions():
    # NOTE(andreykurilin): available clients version should not be
    # hardcoded, so let's discover them.
    matcher = re.compile(r"v[0-9]*$")
    submodules = pkgutil.iter_modules([os.path.dirname(__file__)])
    available_versions = [name[1:] for loader, name, ispkg in submodules
                          if matcher.search(name)]

    return available_versions


def check_major_version(api_version):
    """Checks major part of ``APIVersion`` obj is supported.

    :raises novaclient.exceptions.UnsupportedVersion: if major part is not
                                                      supported
    """
    available_versions = get_available_major_versions()
    if (not api_version.is_null() and
            str(api_version.ver_major) not in available_versions):
        if len(available_versions) == 1:
            msg = _("Invalid client version '%(version)s'. "
                    "Major part should be '%(major)s'") % {
                "version": api_version.get_string(),
                "major": available_versions[0]}
        else:
            msg = _("Invalid client version '%(version)s'. "
                    "Major part must be one of: '%(major)s'") % {
                "version": api_version.get_string(),
                "major": ", ".join(available_versions)}
        raise exceptions.UnsupportedVersion(msg)


def get_api_version(version_string):
    """Returns checked APIVersion object"""
    version_string = str(version_string)
    if strutils.is_int_like(version_string):
        version_string = "%s.0" % version_string

    api_version = APIVersion(version_string)
    check_major_version(api_version)
    return api_version


def _get_server_version_range(client):
    version = client.versions.get_current()

    if not hasattr(version, 'version') or not version.version:
        return APIVersion(), APIVersion()

    return APIVersion(version.min_version), APIVersion(version.version)


def discover_version(client, requested_version):
    """Discover most recent version supported by API and client.

    Checks ``requested_version`` and returns the most recent version
    supported by both the API and the client.

    :param client: client object
    :param requested_version: requested version represented by APIVersion obj
    :returns: APIVersion
    """
    server_start_version, server_end_version = _get_server_version_range(
        client)

    if (not requested_version.is_latest() and
            requested_version != APIVersion('2.0')):
        if server_start_version.is_null() and server_end_version.is_null():
            raise exceptions.UnsupportedVersion(
                _("Server doesn't support microversions"))
        if not requested_version.matches(server_start_version,
                                         server_end_version):
            raise exceptions.UnsupportedVersion(
                _("The specified version isn't supported by server. The valid "
                  "version range is '%(min)s' to '%(max)s'") % {
                    "min": server_start_version.get_string(),
                    "max": server_end_version.get_string()})
        return requested_version

    if requested_version == APIVersion('2.0'):
        if (server_start_version == APIVersion('2.1') or
                (server_start_version.is_null() and
                 server_end_version.is_null())):
            return APIVersion('2.0')
        else:
            raise exceptions.UnsupportedVersion(
                _("The server isn't backward compatible with Nova V2 REST "
                  "API"))

    if server_start_version.is_null() and server_end_version.is_null():
        return APIVersion('2.0')
    elif novaclient.API_MIN_VERSION > server_end_version:
        raise exceptions.UnsupportedVersion(
            _("Server version is too old. The client valid version range is "
              "'%(client_min)s' to '%(client_max)s'. The server valid version "
              "range is '%(server_min)s' to '%(server_max)s'.") % {
                  'client_min': novaclient.API_MIN_VERSION.get_string(),
                  'client_max': novaclient.API_MAX_VERSION.get_string(),
                  'server_min': server_start_version.get_string(),
                  'server_max': server_end_version.get_string()})
    elif novaclient.API_MAX_VERSION < server_start_version:
        raise exceptions.UnsupportedVersion(
            _("Server version is too new. The client valid version range is "
              "'%(client_min)s' to '%(client_max)s'. The server valid version "
              "range is '%(server_min)s' to '%(server_max)s'.") % {
                  'client_min': novaclient.API_MIN_VERSION.get_string(),
                  'client_max': novaclient.API_MAX_VERSION.get_string(),
                  'server_min': server_start_version.get_string(),
                  'server_max': server_end_version.get_string()})
    elif novaclient.API_MAX_VERSION <= server_end_version:
        return novaclient.API_MAX_VERSION
    elif server_end_version < novaclient.API_MAX_VERSION:
        return server_end_version


def update_headers(headers, api_version):
    """Set microversion headers if api_version is not null"""

    if not api_version.is_null():
        version_string = api_version.get_string()
        if api_version.ver_minor != 0:
            headers[LEGACY_HEADER_NAME] = version_string
        if api_version.ver_minor >= 27:
            headers[HEADER_NAME] = '%s %s' % (SERVICE_TYPE, version_string)


def check_headers(response, api_version):
    """Checks that microversion header is in response."""
    if api_version.ver_minor > 0:
        if (api_version.ver_minor < 27 and
                LEGACY_HEADER_NAME not in response.headers):
            _warn_missing_microversion_header(LEGACY_HEADER_NAME)
        elif (api_version.ver_minor >= 27 and
                HEADER_NAME not in response.headers):
            _warn_missing_microversion_header(HEADER_NAME)


def _add_substitution(versioned_method):
    _SUBSTITUTIONS.setdefault(versioned_method.name, [])
    _SUBSTITUTIONS[versioned_method.name].append(versioned_method)


def _get_function_name(func):
    # NOTE(andreykurilin): Based on the facts:
    #  - Python 2 does not have __qualname__ property as Python 3 has;
    #  - we cannot use im_class here, since we need to obtain name of
    #    function in `wraps` decorator during class initialization
    #    ("im_class" property does not exist at that moment)
    #  we need to write own logic to obtain the full function name which
    #  include module name, owner name(optional) and just function name.
    filename, _lineno, _name, line = traceback.extract_stack()[-4]
    module, _file_extension = os.path.splitext(filename)
    module = module.replace("/", ".")
    if module.endswith(func.__module__):
        return "%s.[%s].%s" % (func.__module__, line, func.__name__)
    else:
        return "%s.%s" % (func.__module__, func.__name__)


def get_substitutions(func_name, api_version=None):
    if hasattr(func_name, "__id__"):
        func_name = func_name.__id__

    substitutions = _SUBSTITUTIONS.get(func_name, [])
    if api_version and not api_version.is_null():
        return [m for m in substitutions
                if api_version.matches(m.start_version, m.end_version)]
    return sorted(substitutions, key=lambda m: m.start_version)


# FIXME(mriedem): This breaks any ManagerWithFind.list method that has a
# 'detailed' kwarg since the ManagerWithFind.findall won't find the correct
# argspec from the wrapped list method.
def wraps(start_version, end_version=None):
    start_version = APIVersion(start_version)
    if end_version:
        end_version = APIVersion(end_version)
    else:
        end_version = APIVersion("%s.latest" % start_version.ver_major)

    def decor(func):
        func.versioned = True
        name = _get_function_name(func)

        versioned_method = VersionedMethod(name, start_version,
                                           end_version, func)
        _add_substitution(versioned_method)

        @functools.wraps(func)
        def substitution(obj, *args, **kwargs):
            methods = get_substitutions(name, obj.api_version)

            if not methods:
                raise exceptions.VersionNotFoundForAPIMethod(
                    obj.api_version.get_string(), name)
            return methods[-1].func(obj, *args, **kwargs)

        # Let's share "arguments" with original method and substitution to
        # allow put utils.arg and wraps decorators in any order
        if not hasattr(func, 'arguments'):
            func.arguments = []
        substitution.arguments = func.arguments

        # NOTE(andreykurilin): The way to obtain function's name in Python 2
        #   bases on traceback(see _get_function_name for details). Since the
        #   right versioned method is used in several places, one object
        #   can have different names. Let's generate name of function one time
        #   and use __id__ property in all other places.
        substitution.__id__ = name

        return substitution

    return decor


def _warn_missing_microversion_header(header_name):
    """Log a warning about missing microversion response header."""
    LOG.warning(_(
        "Your request was processed by a Nova API which does not support "
        "microversions (%s header is missing from response). "
        "Warning: Response may be incorrect."), header_name)


def deprecated_after(version):
    decorator = wraps('2.0', version)

    def wrapper(fn):
        @functools.wraps(fn)
        def wrapped(*a, **k):
            decorated = decorator(fn)
            if hasattr(fn, '__module__'):
                mod = fn.__module__
            else:
                mod = a[0].__module__
            warnings.warn('The %s module is deprecated '
                          'and will be removed.' % mod, DeprecationWarning)
            return decorated(*a, **k)
        return wrapped
    return wrapper
